********************************************************************************
* BiMOMA example code
* Kim J, Reed JL, Maravelias CT (2011) Large-Scale Bi-Level Strain Design
* Approaches and Mixed-Integer Programming Solution Techniques. PLoS ONE 6(9):
* e24162. doi:10.1371/journal.pone.0024162
*
* Joohnoon Kim, Jennifer L. Reed, and Christos T. Maravelias
* Department of Chemical and Biological Engineering
* University of Wisconsin-Madison, Madison, Wisconsin, USA
********************************************************************************

$Offlisting
$include EcoliCoreTextbookModel.gms

$onecho > cplex.op9
indic vbound(jk)$d(jk) 1
indic u2bound(jk)$d(jk) 0
indic u3bound(jk)$d(jk) 1
eprhs 1e-8
epopt 1e-8
epint 0
threads 4
mipstart 1
repairtries 10
rinsheur 100
barqcpepcomp 1e-4
$offecho

Sets
soln             possible solutions in the solution pool /s1*s5/
store(soln)      identified solutions in the solution pool
exchg(j)         used to define exchange reactions /EX_ac_e,EX_akg_e,EX_co2_e,EX_etoh_e,EX_for_e,EX_fum_e,EX_glc_e,EX_h_e,EX_h2o_e,EX_lacD_e,EX_o2_e,EX_pi_e,EX_pyr_e,EX_succ_e/
essen(j)         used to define essential reactions /EX_glc_e,EX_h_e,EX_pi_e,ACONT,Biomass,CS,ENO,GAPD,GLCpts,ICDHyr,PGK,PGM,PIt,RPI/
nouse(j)         used to define unused reactions /EX_fum_e,FUMt2_2/
trans(j)         used to define transport reactions /ACt2r,AKGt2r,CO2t,DLACt2,ETOHt2r,FORt,FUMt2_2,GLCpts,H2Ot,O2t,PIt,PYRt2r,SUCCt2_2,SUCCt2b/
exclude(j)       used to define excluded reactions /ATPM,ATPS4r,ADK1,TKT1,TKT2/
bnd(j)           used to define bounded reactions
jk(j)            used to define reactions considered for deletion
subi(i)          used to define subset of metabolites
subj(j)          used to define subset of reactions
momaset(j)       used to define subset of reactions considered in moma objective;

*jk(j)=not (exchg(j) or essen(j) or nouse(j) or trans(j) or exclude(j));
jk(j)=not (exchg(j) or essen(j) or nouse(j) or trans(j));
subj(j)=not nouse(j);
subi(i)$sum(subj$S(i,subj),1)=yes;
momaset(j)= subj(j) - exchg(j);

Scalar
alpha            minimum increase of the objective per deletion
gamma            penalty for dual variables
Dmax             maximum number of deletions
Umax             bounds on the dual variables /100/
tolerance /1e-12/;

Parameters
c(j)             used to define the objective function for outer problem
pri(j)           used to define the objective function for inner problem
v_wt(j)          wildtype flux distribution
storeKnock(soln) store number of reaction deletions for each solution
storeD(j,soln)   store reaction deletions for each solution
storeV(j,soln)   store reaction flux value for each solution
maxobj(soln)     maximum objective value for each solution
MOMAprod(soln)   minimum euclidean distance production
MOMAgrow(soln)   minimum euclidean distance growth;

Variables
v(j)             flux values through reaction in network
u1(i)            dual variable for S*v = 0
u2(j)            dual variable for   v = 0
Obj              the objective function
distance         minimum euclidean distance objective;

Positive Variables
u3(j)            dual variable for -v =l= -vmin;

Binary Variables
d(j)             indicator for reactions;

Equation
calcobj          objective function
optobj           BiMOMA objective function
eucl_distance    euclidean distance objective function
maxknock         maximum number of allowed knockouts
mingrowth        minimum growth requirement
primaldual       primal objective = dual objective
massbalance(i)   steady state mass balance
dualbalance(j)   dual constraint for primal variable v
vbound(j)        set bound for v from reaction indicators
u2bound(j)       set bound for u2 from reaction indicators
u3bound(j)       set bound for u3 from reaction indicators
intcut(soln)     integer cut to generate diverse solutions;

calcobj..        Obj=e=sum( j$subj(j),c(j)*v(j) );
optobj..         Obj=e=sum( j$subj(j),c(j)*v(j) ) - alpha*sum(jk, d(jk)) - gamma*sum(jk, sqr(u2(jk)));
eucl_distance..  distance=e=sum(j$momaset(j), sqr(v_wt(j)-v(j)));
maxknock..       sum( jk,d(jk) )=l=Dmax;
mingrowth..      sum( j$subj(j),pri(j)*v(j) )=g=0.1;

massbalance(i)$subi(i).. sum( j$subj(j),S(i,j)*v(j) )=e=0;
dualbalance(j)$subj(j).. sum( i$subi(i),S(i,j)*u1(i) )+u2(j)$jk(j)+u3(j)$bnd(j)=e=(v(j)-v_wt(j))$momaset(j);
primaldual..             sum( j$subj(j),(sqr(v(j))-v_wt(j)*v(j))$momaset(j)-(LowerLimits(j)*u3(j))$bnd(j))=l=0;
vbound(jk)..             v(jk)=e=0;
u2bound(jk)..            u2(jk)=e=0;
u3bound(jk)$bnd(jk)..    u3(jk)=e=0;

intcut(soln)$store(soln).. sum(jk$storeD(jk,soln),d(jk))-sum(jk$(not storeD(jk,soln)),d(jk))
                           =l=storeKnock(soln)-1;

Model FBA    /calcobj,massbalance/;
Model MOMA   /eucl_distance,massbalance/;
Model BiMOMA /optobj,maxknock,mingrowth,primaldual,massbalance,dualbalance,
              vbound,u2bound,u3bound,intcut/;

option miqcp=cplex;

BiMOMA.optfile=9;
BiMOMA.optcr=0;
BiMOMA.holdfixed=1;
BiMOMA.reslim=1000000;

* Set the lower and upper bounds for variables
LowerLimits('ATPM')=7.6;
LowerLimits('EX_o2_e')=-Vmax;
LowerLimits('EX_co2_e')=-Vmax;
LowerLimits('EX_h_e')=-Vmax;
LowerLimits('EX_h2o_e')=-Vmax;
LowerLimits('EX_pi_e')=-Vmax;
LowerLimits('EX_glc_e')=-Vmax;

bnd(j)$(LowerLimits(j) gt -Vmax)=yes;
LowerLimits(j)$(LowerLimits(j) = -Vmax)=-inf;
UpperLimits(j)=inf;

v.lo(j)=LowerLimits(j);
v.up(j)=UpperLimits(j);

* Maximize wildtype growth rate
v.lo('EX_glc_e')=-10;
c('Biomass')=1;
solve FBA using lp maximizing Obj;
* Store wildtype flux distributions
v_wt(j)=v.l(j);

* Define the outer and inner objectives
c('Biomass')=0;
pri('Biomass')=1;
c('EX_pyr_e')=1;

* Calculate the theoretical maximum production
solve FBA using lp maximizing Obj;
abort$(Obj.l=0) "Maximum production rate is zero";
* Set the minimum increase of the objective per deletion
alpha=0.1;
gamma=1e-6;

* Set the lower and upper bounds for variables
u2.lo(j)=-Umax;  u2.up(j)=Umax;

* Define the allowed number of knockouts
Dmax=2;

* Initialize parameter values
storeKnock(soln)=Dmax;   storeD(j,soln)=0;        storeV(j,soln)=0;
maxobj(soln)=0;          MOMAprod(soln)=0;        MOMAgrow(soln)=0;

scalar iter /1/;
file result /BiMOMA_CoreNetwork.txt/;
result.pw=1000; result.pc=5; result.ps=130;
put result;
put 'Number';
put "Deletions";
while(iter lt Dmax, put ''; iter=iter+1;);
put 'Growth';
put 'MaxProd';
put 'MOMA_growth';
put 'MOMA_Prod';
put /;
putclose result;
result.ap=1;

Alias(soln,temp);
store(soln)=no;

loop(temp,
         v.lo(j)=LowerLimits(j);
         v.up(j)=UpperLimits(j);

         solve BiMOMA using miqcp maximizing Obj;

         store(temp)$((BiMOMA.modelstat = 1) or (BiMOMA.modelstat = 8))=yes;
         if (store(temp),
         storeKnock(temp)=sum(jk, d.l(jk));
         storeD(j,temp)=d.l(j);
         storeV(j,temp)=v.l(j);
         maxobj(temp)=sum(j$c(j), v.l(j));

         v.fx(jk)$d.l(jk)=0;
         v.lo(j)$(abs(v.lo(j)) lt tolerance) = 0;
         v.up(j)$(abs(v.up(j)) lt tolerance) = 0;

         solve MOMA using qcp minimizing distance;
         MOMAprod(temp)=sum(j$c(j), v.l(j));
         MOMAgrow(temp)=sum(j$pri(j), v.l(j));
         v.l(j)=storeV(j,temp);

*This section writes out the results to a txt file
         iter=1;
         put result;
         put storeKnock(temp);
         loop(jk$storeD(jk,temp), put jk.tl; );
         iter=storeKnock(temp);
         while(iter lt Dmax, put ''; iter=iter+1;);
         put storeV('Biomass',temp):8:4;
         put maxobj(temp):8:4;
         put MOMAgrow(temp):8:4;
         put MOMAprod(temp):8:4;
         put /;
         putclose result;
         );
);
